//#pragma config |= 0x3E50
//#pragma origin 4

#define cls instruction(00000001b); cursorpos=0x00;
#define numcols 8
#define lastcol (numcols - 1)
#define delay 2
#define _4MHZ

#include <system.h>

volatile bit db0 @ PORTA.6;
volatile bit db1 @ PORTB.7;
volatile bit db2 @ PORTB.6;
volatile bit db3 @ PORTB.4;
volatile bit db4 @ PORTA.2;
volatile bit db5 @ PORTA.3;
volatile bit db6 @ PORTA.4;
volatile bit db7 @ PORTB.0;
bit RW @ PORTA.0;
bit RS @ PORTA.1;
bit E @ PORTA.7;
bit backlight @ PORTB.3;
bit SCROLL, WRAP, DD, CC, BB;
char data_in, cursorpos, temp;
char cmdbuffer[40], *cmdbufferindex; 
//command buffer operation:
//pointer (index) points to NEXT 'EMPTY' location in the array.

//#pragma rambank 1
char textbuffer[40], *textbufferindex;
//text buffer is used only for shifting a row
//entire row is read into this array, 
//then written back to display in the new row
//#pragma rambank 0

void instruction(char inst);
void printf(char character);
void setout(char byte);
void MC(); //move cursor
void dispooc(); //display on/off control
void shiftrow2up();
char read();
void back();
char readaddress();

volatile bit go @ ADCON0.GO;
volatile bit gie @ INTCON.GIE;
volatile bit rcie @ PIE1.RCIE;
volatile bit peie @ INTCON.PEIE;
volatile bit rcif @ PIR1.RCIF;
volatile bit trmt @ TXSTA.TRMT;

void interrupt(void)
{
	if(rcif) //UART Recieve Interrupt
    {
        if(cmdbufferindex < (&cmdbuffer[0]+40))
        {
            *cmdbufferindex = rcreg;
            cmdbufferindex++;
        }
        rcreg = 0;
    }
}

//#include "d:\pic\programs\headers\delay.h"

void instruction(char inst)
{
    RS=0;
    RW=0;
    setout(inst);
    delay_10us(delay);
    E=1;
    E=0;
    delay_10us(delay);    
}

char read()
{
    char value;
    RS=1;
    RW=1;
    trisa = 11111111b;
    trisb = 11000111b;
    E=1;
    delay_10us(delay);
    E=0;
    delay_10us(delay);
    value=db7;
    value*=2;
    value+=db6;
    value*=2;
    value+=db5;
    value*=2;
    value+=db4;
    value*=2;
    value+=db3;
    value*=2;
    value+=db2;
    value*=2;
    value+=db1;
    value*=2;
    value+=db0;
    trisa = 00100000b; //0 = Output, 1 = Input
    trisb = 11000110b; //original settings (restored)
    return value;
}

char readaddress()
{
    char address;
    RS=0;
    RW=1;
    trisa = 11111111b;
    trisb = 11000111b;
    E=1;
    delay_10us(delay);
    E=0;
    while(db7==1){}
    address=db6;
    address*=2;
    address+=db5;
    address*=2;
    address+=db4;
    address*=2;
    address+=db3;
    address*=2;
    address+=db2;
    address*=2;
    address+=db1;
    address*=2;
    address+=db0;
    trisa = 00100000b; //0 = Output, 1 = Input
    trisb = 11000110b; //original settings (restored)
    return address;
}

void printf(char character)
{
    RS=1;
    RW=0;
    setout(character);
    delay_10us(delay);
    E=1;
    delay_10us(delay);
    E=0;
    delay_10us(delay);
    if(WRAP)
    {
        if((cursorpos>=lastcol) && (cursorpos < 0x27)) //at end of first line
        {    
            cursorpos=0x40;
            MC();
        }
        else
        { 
            if(cursorpos >= (lastcol + 0x40)) //at end of second line
            {
                if(SCROLL==0)
                {
                    cursorpos=0x00;
                    MC();
                }
                else
                {
                    shiftrow2up();
                    cursorpos=0x40;
                    MC();
                }
            }
            else              
                cursorpos+=1;
        }
    }
    else if(cursorpos == 0x67)
        cursorpos=0x00;
    else
        cursorpos += 1;
}

void shiftrow2up()
{
    cursorpos=0x40;
    MC();
    char i;
    for(i=0;i<=40;i++)
    {
        temp=read();
        cursorpos=i;        
        MC();
       
        RS=1;           //can't use printf
        RW=0;
        setout(temp);
        delay_10us(delay);
        E=1;
        E=0;
        delay_10us(delay);
        cursorpos+=0x40;
        MC();
    }
    cursorpos=0x40;
    MC();
}

void setout(char byte)
{
    db0=(test_bit(byte,0)!=0);
    db1=(test_bit(byte,1)!=0);
    db2=(test_bit(byte,2)!=0);
    db3=(test_bit(byte,3)!=0);
    db4=(test_bit(byte,4)!=0);
    db5=(test_bit(byte,5)!=0);
    db6=(test_bit(byte,6)!=0);
    db7=(test_bit(byte,7)!=0);
}

void MC() //move cursor to position specified by cursorpos
{
    instruction(10000000b + cursorpos);
}

void dispooc() //display on/off control
{
    char insttemp;
    /*insttemp = 00001000b;
    insttemp.2 = DD;
    insttemp.1 = CC;
    insttemp.0 = BB;*/
    insttemp=2;
    insttemp+=DD;
    insttemp*=2;
    insttemp+=CC;
    insttemp*=2;
    insttemp+=BB;
    instruction(insttemp);
}

void init()
{
    gie=0;
    delay_ms(500);
    instruction(0x30);
    delay_ms(6);
    instruction(0x30);
    delay_ms(1);
    instruction(0x30);
    delay_ms(1);
    instruction(00111000b);
    delay_ms(1);
    instruction(00001000b);
    delay_ms(1);
    instruction(00000001b);
    delay_ms(1);
    instruction(00000110b); //cursor moves L to R (no display shift)
    delay_ms(1);
    instruction(00001111b); //display on, cursor on, cursor blinking on
    gie=1;
}

void back()
{
    if(cursorpos==0x00)
    {
        if(WRAP==1)
            cursorpos = (lastcol + 0x40);
        else
            cursorpos = 0x67;
    }
    else if(cursorpos==0x40)
    {
        if(WRAP==1)
            cursorpos=lastcol;
        else
            cursorpos=0x27;
    }
    else
        cursorpos--;
    MC();
}

void main()
{
    cmdbufferindex = &cmdbuffer[0];
    //VRCON = 10001000b; //int Vref on, set to initial value of 1/2 Vdd
    osccon =01100000b; //internal oscillator @ 4MHz
    ansel= 	00000000b; //all A/D off
    cmcon = 00000111b; //comparators off
    porta = 00000000b;  
    trisa = 00100000b; //0 = Output, 1 = Input
    portb = 00000000b;
    trisb = 00000110b;
    spbrg=12; //19200 baud @ 4MHz @ high speed
    //SPBRG=0x19; //9600 baud @ 4MHz @ high speed
    txsta=	00100100b; //full duplex asynchronous high speed
    //SPBRG=0x19; //2400 baud @ 4MHz @ low speed    
    //TXSTA=0b.0010.0000; //full duplex asynchronous low speed
    rcsta =	10010000b; //full duplex asynchronous 
    clear_bit(option_reg,7); //PORTB weak pull-ups enabled
    init();
    cursorpos=0x00;
    DD=1;
    CC=1;
    BB=1;
    dispooc();
    backlight=1;
    WRAP=1;
    SCROLL=0;
    rcie = 1;
    peie = 1;

    cls
    while(1)
    {
        if(cmdbufferindex > &cmdbuffer[0])
        {
            rcie=0;
            cmdbufferindex--;
            temp=*cmdbufferindex;
            if(temp<32)
            {
                switch (temp) 
                {
                    case 0x01: //CTRL+A - cursor home
                        instruction(00000010b);
                        break;
                    case 0x02: //CTRL+B - hide display
                        DD=0;
                        dispooc();
                        break;
                    case 0x03: //CTRL+C - restore display
                        DD=1;
                        dispooc();
                        break;
                    case 0x04: //CTRL+D - hide cursor
                        BB=0;
                        CC=0;
                        dispooc();
                        break;
                    case 0x05: //CTRL+E - show underline cursor
                        CC=1;
                        dispooc();
                        break;
                    //case 0x06: //CTRL+F - show block cursor
                        //break;
                    //case 0x07: //CTRL+G - show inverting block cursor
                        //break;
                    case 0x08: //CTRL+H - backspace (destructive)
                        back();
                        printf(' ');
                        back();                    
                        break;
                    //case 0x09: //CTRL+I - control the boot screen
                        //break;
                    case 0x0A: //CTRL+J - line feed
                        if(cursorpos < 0x40) //line 1
                        {
                            cursorpos = cursorpos + 0x40;
                            MC();                       
                        }
                        else if(SCROLL == 0)
                        {
                            cursorpos = cursorpos - 0x40;
                            MC();
                        }
                        else
                        {
                            shiftrow2up();
                        }
                        break;
                    case 0x0B: //CTRL+K - delete in place
                        printf(' ');
                        back();
                        break;
                    case 0x0C: //CTRL+L - clear display
                        cls break;
                    case 0x0D: //CTRL+M - carriage return
                        if(cursorpos<0x40)
                            cursorpos=0x00;
                        else
                            cursorpos=0x40;
                        MC();
                        break;
                    case 0x0E: //CTRL+N - backlight control
                        cmdbufferindex++; //shift to first empty position
                        temp=(char)cmdbufferindex;
                        rcie=1;
                        while(cmdbufferindex<(temp+1)){} //receive 1 character
                        rcie=0;
                        cmdbufferindex-=1;
                        if(*cmdbufferindex!=0)
                            backlight=1;
                        else
                            backlight=0;
                        cmdbufferindex--;
                        break;
                    case 0x11: //CTRL+Q - set cursor position
                        cmdbufferindex++; //shift to first empty position
                        temp=(char)cmdbufferindex;
                        rcie=1;
                        while(cmdbufferindex < (temp+2)){} //recieve two characters
                        rcie=0;
                        cmdbufferindex-=2;
                        cursorpos = *cmdbufferindex; //set column
                        cmdbufferindex++;
                        if(*cmdbufferindex==1)
                            cursorpos += 0x40;
                        MC();
                        cmdbufferindex -=2;
                        break;
                    case 0x13: //CTRL+S - SCROLL ON
                        //SCROLL=1; //disabled until scroll code finished
                        break;
                    case 0x14: //CTRL+T - SCROLL OFF
                        SCROLL=0;
                        break;
                    case 0x15: //CTRL+U - set scrolling marquee characters
                        break;
                    case 0x16: //CTRL+V - enable scrolling marquee
                        break;
                    case 0x17: //CTRL+W - WRAP ON
                        WRAP=1;
                        break;
                    case 0x18: //CTRL+X - WRAP OFF
                        WRAP=0;
                        break;
                    case 0x19: //CTRL+Y - set custom char bitmap
                        break;
                    case 0x1A: //CTRL+Z - reboot
                        init();
                        break;
                    case 0x1B: //escape - escape sequence prefix
                        break;
                    case 0x1C: //CTRL+EQUAL - send data directly to LCD controller
                        cmdbufferindex++; //shift to first empty position
                        temp=(char)cmdbufferindex;
                        rcie=1;
                        while(cmdbufferindex < (temp+2)){} //recieve two characters
                        rcie=0;
                        cmdbufferindex-=2;
                        temp=*cmdbufferindex; //location
                        cmdbufferindex++;
                        if(temp==0)
                        {
                            instruction(*cmdbufferindex);
                        }
                        else if(temp==1)
                        {
                            RS=1;
                            RW=0;
                            setout(*cmdbufferindex);
                            //delay_10us(3);
                            E=1;
                            delay_10us(1);
                            E=0;
                            delay_10us(3);
                        }
                        //else if(temp==2){} //used for 2-controller displays?
                        cmdbufferindex -=2;
                        break;
                    default: 
                        break;
                }
            }
            else
                printf(*cmdbufferindex);
            rcie=1;
        }
    }
}